#!/usr/local/bin/ruby

require 'getoptlong'

class ConffileReader
	attr_reader :configuration

	def initialize(filename, cascade=false)
		if(!filename.respond_to?(:to_s))
			raise TypeError, "filename must respond to to_s()"
		end
		
		if(cascade != true and cascade != false)
			raise TypeError, "cascade must be either true or false"
		end
		
		@cascade = cascade
		@filename = filename.to_s()
		@configuration = Configuration.new()
		
		readconfiguration()
		
	end

protected

	def readconfiguration()
		log("opening #{@filename} to read configuration", 5)
		File.open(@filename, File::RDONLY) { |fp|
			
			inchunk = false
			monitorvars = Hash.new()
			clearmonitorvars(monitorvars)
			
			line = fp.gets()
			while(line != nil or inchunk == true)
				line = line.to_s().strip()
				
				# drop any comment lines
				if(line.length > 0 and line[0].chr() == '#')
					# do nothing
				elsif(inchunk == false and line != '')
					inchunk = true
				elsif(inchunk == true and line == '')
					if(monitorvars['checkregex'] =~ /^\/.*\/$/)
						monitorvars['checkregex'] = Regexp.new(monitorvars['checkregex'][/^\/(.*)\/$/,1])
					else
						monitorvars['checkregex'] = Regexp.new(monitorvars['checkregex'])
					end
					if(monitorvars['checkinterval'] != nil)
						@configuration << MonitorConfiguration.new(monitorvars['name'], monitorvars['checkcommand'], monitorvars['checkregex'], monitorvars['onfailcommand'], monitorvars['alertcommand'], monitorvars['checkinterval'])
					elsif(monitorvars['checkinterval'] and monitorvars['aftercommandwait'] != nil)
						@configuration << MonitorConfiguration.new(monitorvars['name'], monitorvars['checkcommand'], monitorvars['checkregex'], monitorvars['onfailcommand'], monitorvars['alertcommand'], monitorvars['checkinterval'], monitorvars['aftercommandwait'])
					else
						@configuration << MonitorConfiguration.new(monitorvars['name'], monitorvars['checkcommand'], monitorvars['checkregex'], monitorvars['onfailcommand'], monitorvars['alertcommand'])
					end
					inchunk = false
					if(@cascade == false)
						log("clearing variables before reading next configuration section - cascading is set to false", 8)
						clearmonitorvars(monitorvars)
					end
				end
				
				if(inchunk == true)
					# split the line to get the variable being set and the value
					begin
						command = line[/([^:]+):/, 1]
						if(command == nil)
							raise Exception, "malformed line read in configuration at line #{fp.lineno} - no setting name found"
						end
						
						setting = line[/([^:]+):(.*)/,2].strip()
						if(setting == nil)
							raise Exception, "malformed line read in configuration at line #{fp.lineno} - no setting found"
						end
						
						if(!monitorvars.has_key?(command))
							raise Exception, "malformed line read in configuration at line #{fp.lineno} - setting name #{command} unknown"
						end
					rescue Exception
						log("#{$!}")
						exit!(1)
					end
					monitorvars[command] = setting
				end
				
				# get the next line
				line = fp.gets()
			end
		}
		log("closing #{@filename} after reading configuration", 5)
	end
	
	
	# helper function to clear the monitor variable hash - called by readconfiguration
	def clearmonitorvars(monitorvars)
		monitorvars['name'] = nil
		monitorvars['checkcommand'] = nil
		monitorvars['checkregex'] = nil
		monitorvars['onfailcommand'] = nil
		monitorvars['checkinterval'] = nil
		monitorvars['aftercommandwait'] = nil
		monitorvars['alertcommand'] = nil
	end
end


class Configuration < Array
end


class MonitorConfiguration
	attr_reader :name, :checkcommand, :checkregex, :onfailcommand, :alertcommand, :checkinterval, :aftercommandwait
	
	def initialize(name, checkcommand, checkregex, onfailcommand, alertcommand, checkinterval=30, aftercommandwait=5)
		if(!name.respond_to?(:to_s))
			raise TypeError, "name must respond to to_s()"
		end
		
		if(!checkcommand.respond_to?(:to_s))
			raise TypeError, "checkcommand must respond to to_s()"
		end
		
		if(!checkregex.kind_of?(Regexp))
			raise TypeError, "checkregex must be a regular expression"
		end
		
		if(!onfailcommand.respond_to?(:to_s))
			raise TypeError, "onfailcommand must respond to to_s()"
		end
		
		if(!checkinterval.respond_to?(:to_i))
			raise TypeError, "checkinterval must respond to to_i()"
		end
		
		if(!aftercommandwait.respond_to?(:to_i))
			raise TypeError, "aftercommandwait must respond to to_i()"
		end
		
		if(alertcommand != nil and !alertcommand.respond_to?(:to_s))
			raise TypeError, "alertcommand must respond to to_s()"
		end
		
		@name = name.to_s()
		@checkcommand = checkcommand.to_s()
		@checkregex = checkregex
		@onfailcommand = onfailcommand.to_s()
		@alertcommand = alertcommand.to_s()
		@checkinterval = checkinterval.to_i()
		@aftercommandwait = aftercommandwait.to_i()
		
		log("Read conifguration section:\n#{to_s()}", 8)
	end
	
	
	def to_s()
		output = <<OUTPUT
name - #{@name}
checkcommand - #{@checkcommand}
checkregex - #{@checkregex.source}
onfailcommand - #{@onfailcommand}
alertcommand - #{@alertcommand}
checkinterval - #{@checkinterval}
aftercommandwait - #{@aftercommandwait}
OUTPUT

		return output
	end
end


class MonitorGroup

	def initialize()
		@monitors = Array.new()
	end
	
	def createmonitor(configuration)
		if(!configuration.kind_of?(MonitorConfiguration))
			raise TypeError, "configuration must be a MonitorConfiguration object"
		end
		
		@monitors << Monitor.new(configuration)
	end
	
	
	def run
		log("Starting all Monitor threads in the group", 5)
		@monitors.each() { |monitor|
			monitor.start()
		}
		
		@monitors.each() { |monitor|
			monitor.join()
		}
	end
	
	
	def stop()
		log("stopping Monitor threads", 5)
		@monitors.each() { |monitor|
			monitor.stop()
		}
	end
end


class Monitor

	def initialize(configuration)
		if(!configuration.kind_of?(MonitorConfiguration))
			raise TypeError, "configuration must be a MonitorConfiguration object"
		end
		@configuration = configuration
		@threaddone = false
	end
	
	
	def start()
		@thread = Thread.new() {
			while(@threaddone == false)
				check()
				sleep(@configuration.checkinterval)
			end
		}
	end
	
	
	def join()
		@thread.join()
	end
	
	
	def stop()
		@threaddone = true
	end
	
protected

	def check()
		log("(#{@configuration.name}) Running #{@configuration.checkcommand}", 5)
		output = `#{@configuration.checkcommand}`
		if(!@configuration.checkregex.match(output))
			log("(#{@configuration.name}) Failed to find \"/#{@configuration.checkregex.source}/\" in output of \"#{@configuration.checkcommand}\"")
			# spawn a thread to run the onfailcommand
			@failcommandrunner = Thread.new() {
				log("(#{@configuration.name}) Running fail command - \"#{@configuration.onfailcommand}\"")
				`#{@configuration.onfailcommand}`
				if(@configuration.alertcommand != '')
					log("(#{@configuration.name}) Running alert command - \"#{@configuration.alertcommand}\"")
					`#{@configuration.alertcommand}`
				end
			}
			
			# wait for aftercommandwait seconds
			sleep(@configuration.aftercommandwait)
		end
	end

end


# prints basic usage
def printusage()
	usage = <<USAGE
Daedalus 1.1 of 17 June 2004

Usage:
	daedalus.rb [options]
Options:
	-h, --help 	Print this message
	-c, --conffile  The location of the configuration file
	-l, --logfile 	The location of the log file-C, --cascadeconfig	Cascade configuration options from one block to the next
	-C, --cascadeconfig	Cascade configuration items from one block to the next
	-N, --nodaemon	Don't detach from the terminal and become a daemon process
	-p, --pidfile	The location of the pid file
	-v, --verbose	Log verbosely for debugging purposes
USAGE

	print usage

end


# forks the program into the background and detaches it from the terminal
def daemonize()
	# fork into the background
	# Parent exits, child continues.
	if(fork())
		exit!()
	end
	
	# Become session leader.
	Process.setsid()
	
	# Zap session leader.
	if(fork())
		exit!()
	end
	
	# Release old working directory.
	Dir.chdir "/"
	
	# Ensure sensible umask. Adjust as needed.
	File.umask 0000
	
	# reopen the file descriptors to nothing
	STDIN.reopen '/dev/null'
	STDOUT.reopen '/dev/null', 'a'
	STDERR.reopen STDOUT
end


# writes a message to the log file
def log(str, level=0)
	if(level == 0)
		File.open($options['logfile'], File::WRONLY|File::CREAT|File::APPEND, 0600) { |fp|
			fp.write("#{Time.now()} - #{str}\n")
		}
	elsif(level > 0 and $options['verbose'] == true)
		File.open($options['logfile'], File::WRONLY|File::CREAT|File::APPEND, 0600) { |fp|
			fp.write("#{Time.now()} - #{str}\n")
		}
	end
end


# writes the pid to a file
def writepid(pidfile)
	File.open(pidfile, File::WRONLY|File::CREAT|File::TRUNC, 0644) { |fp|
		fp.write("#{Process.pid}\n")
	}
end


##### MAIN #####

$options = Hash.new()
$options['conffile'] = 'daedalus.conf'
$options['logfile'] = '/var/log/daedalus.log'
$options['cascadeconfig'] = false
$options['verbose'] = false
$options['daemonize'] = true
$options['pidfile'] = '/var/run/daedalus.pid'

Thread.abort_on_exception = true

# trap kill so we can at least try to cleanup
trap("KILL") {
	log("recieved KILL signal")
	begin
		File.unlink($options['pidfile'])
	rescue Exception
		# do nothing except note that we failed
		$stderr.print "Unable to delete pid file \"#{$options['pidfile']}\" - #{$!}\n"
	end
	exit!(1)
}

begin
	opts = GetoptLong.new(
		['--help', '-h', GetoptLong::NO_ARGUMENT],
		['--conffile', '-c', GetoptLong::REQUIRED_ARGUMENT],
		['--logfile', '-l', GetoptLong::REQUIRED_ARGUMENT],
		['--cascadeconfig', '-C', GetoptLong::NO_ARGUMENT],
		['--nodaemon', '-N', GetoptLong::NO_ARGUMENT],
		['--pidfile', '-p', GetoptLong::REQUIRED_ARGUMENT],
		['--verbose', '-v', GetoptLong::NO_ARGUMENT]
	)
	
	opts.each() { |option,argument|
		if(option == '--help')
			printusage()
			exit!()
		elsif(option == '--conffile')
			$options['conffile'] = argument
		elsif(option == '--logfile')
			$options['logfile'] = argument
		elsif(option == '--cascadeconfig')
			$options['cascadeconfig'] = true
		elsif(option == '--verbose')
			$options['verbose'] = true
		elsif(option == '--nodaemon')
			$options['daemonize'] = false
		elsif(option == '--pidfile')
			$options['pidfile'] = argument
		end
	}
rescue Exception
	$stderr.print "#{$!}\n"
	exit(1)
end

begin
	confreader = ConffileReader.new($options['conffile'], $options['cascadeconfig'])
rescue Exception
	$stderr.print "Unable to read configuration file - #{$!}\n"
	exit(1)
end
monitorgroup = MonitorGroup.new()
confreader.configuration.each() { |monitorconf|
	monitorgroup.createmonitor(monitorconf)
}
if($options['daemonize'] == true)
	daemonize()
end

writepid($options['pidfile'])

log('Daedalus started')

# trap term lower down so we can exit gracefully
trap("TERM") {
	log("recieved TERM signal")
	monitorgroup.stop()
	begin
		File.unlink($options['pidfile'])
	rescue Exception
		# do nothing except note that we failed
		$stderr.print "Unable to delete pid file \"#{$options['pidfile']}\" - #{$!}\n"
		exit(1)
	end
}

monitorgroup.run()
log('Daedalus stopped')

